#!/usr/bin/python
# Licensed under the GNU General Public License Version 2
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.

# Copyright (C) 2008
#    Richard Hughes <richard@hughsie.com>

# imports
from .enums import *
from .package import PackagekitPackage
import collections

class PackagekitFilter(object, PackagekitPackage):

    def __init__(self, fltlist="none"):
        ''' save state '''
        self.fltlist = fltlist
        self.package_list = [] #we can't do emitting as found if we are post-processing
        self.installed_unique = {}

    def add_installed(self, pkgs):
        ''' add a list of packages that are already installed '''
        for pkg in pkgs:
            self.package_list.append((pkg, INFO_INSTALLED))

    def add_available(self, pkgs):
        ''' add a list of packages that are available '''
        for pkg in pkgs:
            self.package_list.append((pkg, INFO_AVAILABLE))

    def add_custom(self, pkg, info):
        ''' add a custom packages indervidually '''
        self.package_list.append((pkg, info))

    def _filter_base(self, pkg):
        ''' do extra filtering (gui, devel etc) '''
        for flt in self.fltlist:
            if flt in (FILTER_GUI, FILTER_NOT_GUI):
                if not self._do_gui_filtering(flt, pkg):
                    return False
            elif flt in (FILTER_DEVELOPMENT, FILTER_NOT_DEVELOPMENT):
                if not self._do_devel_filtering(flt, pkg):
                    return False
            elif flt in (FILTER_FREE, FILTER_NOT_FREE):
                if not self._do_free_filtering(flt, pkg):
                    return False
            elif flt in (FILTER_ARCH, FILTER_NOT_ARCH):
                if not self._do_arch_filtering(flt, pkg):
                    return False
        return True

    def _filter_installed(self, pkg):
        ''' do extra filtering (gui, devel etc) '''
        for flt in self.fltlist:
            if flt in (FILTER_INSTALLED, FILTER_NOT_INSTALLED):
                if not self._do_installed_filtering(flt, pkg):
                    return False
        return True

    def get_package_list(self):
        '''
        do filtering we couldn't do when generating the list
        '''

        # filter common things here like architecture
        # NOTE: we can't do installed and ~installed here as we need
        # this data for the newest and downgrade checks below
        package_list = self.package_list
        self.package_list = []
        for pkg, state in package_list:
            if self._filter_base(pkg):
                self.package_list.append((pkg, state))

        # prepare lookup table of installed packages
        installed_dict = collections.defaultdict(list)
        for pkg, state in self.package_list:
            if state is INFO_INSTALLED:
                installed_dict[self._pkg_get_name(pkg)].append(pkg)

        # check there are not available versions in the package list
        # that are older than the installed version
        package_list = self.package_list
        self.package_list = []
        for pkg, state in package_list:

            add = True
            if state is INFO_AVAILABLE:
                for pkg_tmp in installed_dict[self._pkg_get_name(pkg)]:
                    rc = self._pkg_compare(pkg, pkg_tmp)

                    # don't add if the same as the installed package
                    # or a downgrade to the existing installed package
                    if rc == 0 or rc == -1:
                        add = False
                        break

            if add:
                self.package_list.append((pkg, state))

        # filter installed state last
        package_list = self.package_list
        self.package_list = []
        for pkg, state in package_list:
            if self._filter_installed(pkg):
                self.package_list.append((pkg, state))

        # do the backend specific filtering
        return self.post_process()

    def post_process(self):
        '''
        do filtering we couldn't do when generating the list
        Needed to be implemented in a sub class
        '''
        return self.package_list

    def _pkg_compare(self, pkg1, pkg2):
        '''
        Returns a version comparison of the packages, where:
        -2 : pkg1 not comparable with pkg2
        -1 : pkg2 is newer than pkg1
         0 : pkg1 == pkg2
         1 : pkg1 is newer than pkg2
         2 : not implemented
        Needed to be implemented in a sub class
        '''
        return 2

    def _pkg_get_name(self, pkg):
        '''
        Returns the name of the package used for duplicate filtering
        Needed to be implemented in a sub class
        '''
        return None

    def _pkg_is_installed(self, pkg):
        '''
        Return if the package is installed.
        Needed to be implemented in a sub class
        '''
        return True

    def _pkg_is_devel(self, pkg):
        '''
        Return if the package is development.
        Needed to be implemented in a sub class
        '''
        return True

    def _pkg_is_gui(self, pkg):
        '''
        Return if the package is a GUI program.
        Needed to be implemented in a sub class
        '''
        return True

    def _pkg_is_free(self, pkg):
        '''
        Return if the package is free software.
        Needed to be implemented in a sub class
        '''
        return True

    def _pkg_is_arch(self, pkg):
        '''
        Return if the package is the same architecture as the machine.
        Needed to be implemented in a sub class
        '''
        return True

    def _do_installed_filtering(self, flt, pkg):
        is_installed = self._pkg_is_installed(pkg)
        if flt == FILTER_INSTALLED:
            want_installed = True
        else:
            want_installed = False
        return is_installed == want_installed

    def _do_devel_filtering(self, flt, pkg):
        is_devel = self._pkg_is_devel(pkg)
        if flt == FILTER_DEVELOPMENT:
            want_devel = True
        else:
            want_devel = False
        return is_devel == want_devel

    def _do_gui_filtering(self, flt, pkg):
        is_gui = self._pkg_is_gui(pkg)
        if flt == FILTER_GUI:
            want_gui = True
        else:
            want_gui = False
        return is_gui == want_gui

    def _do_free_filtering(self, flt, pkg):
        is_free = self._pkg_is_free(pkg)
        if flt == FILTER_FREE:
            want_free = True
        else:
            want_free = False
        return is_free == want_free

    def _do_arch_filtering(self, flt, pkg):
        is_arch = self._pkg_is_arch(pkg)
        if flt == FILTER_ARCH:
            want_arch = True
        else:
            want_arch = False
        return is_arch == want_arch




#!/usr/bin/python
#
# Licensed under the GNU General Public License Version 2
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#
# (c) 2008
#     Tim Lauridsen <timlau@fedoraproject.org>

# Misc classes and funtions
import sys

def _isunicode(obj):
    if sys.hexversion >= 0x3000000:
        return isinstance(obj, str)
    else:
        return isinstance(obj, str)

def _israwstring(obj):
    if sys.hexversion >= 0x3000000:
        return isinstance(obj, bytes)
    else:
        return isinstance(obj, str)

def _to_unicode(obj, encoding="utf-8"):
    # string/unicode support
    if _isunicode(obj):
        return obj
    if not _israwstring(obj):
        # don't touch non-str, non-unicode objects
        return obj

    if hasattr(obj, 'decode'):
        return obj.decode(encoding, errors="replace")
    else:
        if sys.hexversion >= 0x3000000:
            return str(obj, encoding, errors="replace")
        else:
            return str(obj, encoding, errors="replace")

def _to_rawstring(obj, from_encoding="utf-8"):
    if _israwstring(obj):
        return obj
    return obj.encode(from_encoding, errors="replace")


class PackageKitPackage:
    '''
    container class from values from the Package signal
    '''
    def __init__(self, info, package_id, summary):
        self.installed = (info == 'installed')
        self.id = _to_rawstring(package_id)
        self.summary = _to_unicode(summary)
        n,v,a,r = self.id.split(';')
        self.name = n
        self.ver = v
        self.arch = a
        self.repoid = r
        self.summary = _to_unicode(summary)
        self.info = _to_rawstring(info)

    def __str__(self):
        (name, ver, arch, repo) = tuple(self.id.split(";"))
        p =  "%s-%s.%s" % (name, ver, arch)
        return p

class PackageKitDistroUpgrade:
    '''
    container class from values from the DistroUpgrade signal
    '''
    def __init__(self, upgrade_type, name, summary):
        self.upgrade_type = upgrade_type
        self.name = _to_rawstring(name)
        self.summary = _to_unicode(summary)

    def __str__(self):
        return " type : %s, name : %s, summary : %s " % (
                self.upgrade_type, self.name, self.summary)

class PackageKitDetails:
    '''
    container class from values from the Detail signal
    '''
    def __init__(self, package_id, package_license, group, detail, url, size):
        self.id = _to_rawstring(package_id)
        self.license = _to_rawstring(package_license)
        self.group = _to_rawstring(group)
        self.detail = _to_unicode(detail)
        self.url = _to_rawstring(url)
        self.size = int(size)

class PackageKitUpdateDetails:
    '''
    container class from values from the UpdateDetail signal
    '''
    def __init__(self, package_id, updates, obsoletes, vendor_url, bugzilla_url, \
                 cve_url, restart, update_text, changelog, state, \
                 issued, updated):
        self.id = _to_rawstring(package_id)
        self.updates = _to_rawstring(updates)
        self.obsoletes = _to_rawstring(obsoletes)
        self.vendor_url = _to_rawstring(vendor_url)
        self.bugzilla_url = _to_rawstring(bugzilla_url)
        self.cve_url = _to_rawstring(cve_url)
        self.restart = (restart == 'yes')
        self.update_text = _to_unicode(update_text)
        self.changelog = _to_unicode(changelog)
        self.state = _to_rawstring(state)
        self.issued = _to_rawstring(issued)
        self.updated = _to_rawstring(updated)

class PackageKitRepos:
    '''
    container class from values from the Repos signal
    '''
    def __init__(self, repo_id, description, enabled):
        self.id = _to_rawstring(repo_id)
        self.description = _to_unicode(description)
        self.enabled = (enabled == 'yes')

class PackageKitFiles:
    '''
    container class from values from the Files signal
    '''
    def __init__(self, package_id, files):
        self.id = _to_rawstring(package_id)
        self.files = _to_rawstring(files)

class PackageKitCategory:
    '''
    container class from values from the Category signal
    '''
    def __init__(self, parent_id, cat_id, name, summary, icon):
        self.parent_id = _to_rawstring(parent_id)
        self.cat_id = _to_rawstring(cat_id)
        self.name = _to_unicode(name)
        self.summary = _to_unicode(summary)
        self.icon = _to_rawstring(icon)

class PackageKitMessage:
    '''container class from values from the Message signal'''
    def __init__(self, code, details):
        self.code = code
        self.details = details



# Licensed under the GNU General Public License Version 2
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#
# Copyright (C) 2008
#    Richard Hughes <richard@hughsie.com>

from packagekit.backend import PackageKitEnum

class PackagekitPackage:

    def get_package_id(self, name, version, arch, data):
        return "%s;%s;%s;%s" % (name, version, arch, data)

    def get_package_from_id(self, package_id):
        ''' split up a package id name;ver;arch;data into a tuple
            containing (name, ver, arch, data)
        '''
        return tuple(package_id.split(';', 4))

    def check_license_field(self, license_field):
        '''
        Check the string license_field for free licenses, indicated by
        their short names as documented at
        http://fedoraproject.org/wiki/Licensing

        Licenses can be grouped by " or " to indicate that the package
        can be redistributed under any of the licenses in the group.
        For instance: GPLv2+ or Artistic or FooLicense.

        Groups of licenses can be grouped with " and " to indicate
        that parts of the package are distributed under one group of
        licenses, while other parts of the package are distributed
        under another group.  Groups may be wrapped in parenthesis.
        For instance:
          (GPLv2+ or Artistic) and (GPL+ or Artistic) and FooLicense.

        At least one license in each group must be free for the
        package to be considered Free Software.  If the license_field
        is empty, the package is considered non-free.
        '''

        groups = license_field.split(" and ")
        f = open('/usr/share/PackageKit/helpers/yum/licenses.txt', 'r')
        free_licenses = f.readlines()

        if len(groups) == 0:
            return False

        one_free_group = False

        for group in groups:
            group = group.replace("(", "")
            group = group.replace(")", "")
            licenses = group.split(" or ")

            group_is_free = False

            for license in licenses:
                license = license.strip()

                if len(license) < 1:
                    continue

                if license in free_licenses:
                    one_free_group = True
                    group_is_free = True
                    break

            if group_is_free == False:
                return False

        if one_free_group == False:
            return False

        return True




# Licensed under the GNU General Public License Version 2
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#
# Copyright (C) 2008
#    Richard Hughes <richard@hughsie.com>

try:
    from collections import Iterable
except ImportError:
    from collections.abc import Iterable


class PackagekitProgress(Iterable):
    '''
    Progress class there controls the total progress of a transaction
    the transaction is divided in n milestones. the class contains
    the current step (milestone n -> n+1) and the percentage of the whole transaction

    Usage:

    from packagekit import PackagekitProgress

    steps = [10, 30, 50, 70] # Milestones in %
    progress = PackagekitProgress(steps)
    for milestone in progress:
        # do the action is this step
        for i in range(100):
            # do some action
            print "progress : %s " % milestone
    '''

    #TODO: Add support for elapsed/remaining time

    def __init__(self, steps=None):
        super(PackagekitProgress, self).__init__()
        if not steps:
            self.reset()
        else:
            self.set_steps(steps)

    def set_steps(self, steps):
        '''
        Set the steps for the whole transaction
        @param steps: list of int representing the percentage of each step in the transaction
        '''
        self.reset()
        self.steps = steps
        self.current_step = 0

    def reset(self):
        self.percent = 0
        self.steps = []
        self.current_step = 0

    def step(self):
        '''
        Step to the next step in the transaction
        '''
        if self.current_step < len(self.steps)-1:
            self.current_step += 1
            self.percent = self.steps[self.current_step]
        else:
            self.current_step = len(self.steps)
            self.percent = 100

    def __iter__(self):
        while self.current_step < len(self.steps):
            yield self.percent
            self.step()

    def _update_percent(self):
        '''
        Increment percentage based on current step
        '''
        if self.current_step == 0:
            startpct = 0
        else:
            startpct = self.steps[self.current_step-1]
        if self.current_step < len(self.steps)-1:
            endpct = self.steps[self.current_step+1]
        else:
            endpct = 100
        incr = endpct -startpct
        self.percent = startpct + incr




# This file was autogenerated from /build/packagekit/src/PackageKit/lib/packagekit-glib2/pk-enum.c by enum-converter.py

class PackageKitEnum:
	exit = ( "unknown", "success", "failed", "cancelled", "key-required", "eula-required", "media-change-required", "killed", "need-untrusted", "cancelled-priority", "skip-transaction", "repair-required", )
	status = ( "unknown", "wait", "setup", "running", "query", "info", "refresh-cache", "remove", "download", "install", "update", "cleanup", "obsolete", "dep-resolve", "sig-check", "test-commit", "commit", "request", "finished", "cancel", "download-repository", "download-packagelist", "download-filelist", "download-changelog", "download-group", "download-updateinfo", "repackaging", "loading-cache", "scan-applications", "generate-package-list", "waiting-for-lock", "waiting-for-auth", "scan-process-list", "check-executable-files", "check-libraries", "copy-files", "run-hook", )
	role = ( "unknown", "cancel", "depends-on", "get-details", "get-details-local", "get-files-local", "get-files", "get-packages", "get-repo-list", "required-by", "get-update-detail", "get-updates", "install-files", "install-packages", "install-signature", "refresh-cache", "remove-packages", "repo-enable", "repo-set-data", "repo-remove", "resolve", "search-details", "search-file", "search-group", "search-name", "update-packages", "what-provides", "accept-eula", "download-packages", "get-distro-upgrades", "get-categories", "get-old-transactions", "repair-system", "upgrade-system", )
	error = ( "unknown", "out-of-memory", "no-cache", "no-network", "not-supported", "internal-error", "gpg-failure", "filter-invalid", "package-id-invalid", "transaction-error", "transaction-cancelled", "package-not-installed", "package-not-found", "package-already-installed", "package-download-failed", "group-not-found", "group-list-invalid", "dep-resolution-failed", "create-thread-failed", "repo-not-found", "cannot-remove-system-package", "process-kill", "failed-initialization", "failed-finalise", "failed-config-parsing", "cannot-cancel", "cannot-get-lock", "no-packages-to-update", "cannot-write-repo-config", "local-install-failed", "bad-gpg-signature", "missing-gpg-signature", "cannot-install-source-package", "repo-configuration-error", "no-license-agreement", "file-conflicts", "package-conflicts", "repo-not-available", "invalid-package-file", "package-install-blocked", "package-corrupt", "all-packages-already-installed", "file-not-found", "no-more-mirrors-to-try", "no-distro-upgrade-data", "incompatible-architecture", "no-space-on-device", "media-change-required", "not-authorized", "update-not-found", "cannot-install-repo-unsigned", "cannot-update-repo-unsigned", "cannot-get-filelist", "cannot-get-requires", "cannot-disable-repository", "restricted-download", "package-failed-to-configure", "package-failed-to-build", "package-failed-to-install", "package-failed-to-remove", "failed-due-to-running-process", "package-database-changed", "provide-type-not-supported", "install-root-invalid", "cannot-fetch-sources", "cancelled-priority", "unfinished-transaction", "lock-required", "repo-already-set", )
	restart = ( "unknown", "none", "system", "session", "application", "security-system", "security-session", )
	filter = ( "unknown", "none", "installed", "~installed", "devel", "~devel", "gui", "~gui", "free", "~free", "visible", "~visible", "supported", "~supported", "basename", "~basename", "newest", "~newest", "arch", "~arch", "source", "~source", "collections", "~collections", "application", "~application", "downloaded", "~downloaded", )
	group = ( "unknown", "accessibility", "accessories", "education", "games", "graphics", "internet", "office", "other", "programming", "multimedia", "system", "desktop-gnome", "desktop-kde", "desktop-xfce", "desktop-dde", "desktop-other", "publishing", "servers", "fonts", "admin-tools", "legacy", "localization", "virtualization", "power-management", "security", "communication", "network", "maps", "repos", "science", "documentation", "electronics", "collections", "vendor", "newest", )
	update_state = ( "unknown", "testing", "unstable", "stable", )
	info = ( "unknown", "installed", "available", "unavailable", "low", "normal", "important", "security", "bugfix", "enhancement", "blocked", "downloading", "updating", "installing", "removing", "cleanup", "obsoleting", "collection-installed", "collection-available", "finished", "reinstalling", "downgrading", "preparing", "decompressing", "untrusted", "trusted", "critical", "install", "remove", "obsolete", "downgrade", )
	sig_type = ( "unknown", "gpg", )
	upgrade = ( "unknown", "stable", "unstable", )
	network = ( "unknown", "offline", "online", "wired", "wifi", "mobile", )
	media_type = ( "unknown", "cd", "dvd", "disc", )
	authorize_type = ( "unknown", "yes", "no", "interactive", )
	upgrade_kind = ( "unknown", "minimal", "default", "complete", )
	transaction_flag = ( "none", "only-trusted", "simulate", "only-download", "allow-reinstall", "just-reinstall", "allow-downgrade", )

# Constants

AUTHORIZE_INTERACTIVE = "interactive"
AUTHORIZE_NO = "no"
AUTHORIZE_UNKNOWN = "unknown"
AUTHORIZE_YES = "yes"
DISTRO_UPGRADE_STABLE = "stable"
DISTRO_UPGRADE_UNKNOWN = "unknown"
DISTRO_UPGRADE_UNSTABLE = "unstable"
ERROR_ALL_PACKAGES_ALREADY_INSTALLED = "all-packages-already-installed"
ERROR_BAD_GPG_SIGNATURE = "bad-gpg-signature"
ERROR_CANCELLED_PRIORITY = "cancelled-priority"
ERROR_CANNOT_CANCEL = "cannot-cancel"
ERROR_CANNOT_DISABLE_REPOSITORY = "cannot-disable-repository"
ERROR_CANNOT_FETCH_SOURCES = "cannot-fetch-sources"
ERROR_CANNOT_GET_FILELIST = "cannot-get-filelist"
ERROR_CANNOT_GET_LOCK = "cannot-get-lock"
ERROR_CANNOT_GET_REQUIRES = "cannot-get-requires"
ERROR_CANNOT_INSTALL_REPO_UNSIGNED = "cannot-install-repo-unsigned"
ERROR_CANNOT_INSTALL_SOURCE_PACKAGE = "cannot-install-source-package"
ERROR_CANNOT_REMOVE_SYSTEM_PACKAGE = "cannot-remove-system-package"
ERROR_CANNOT_UPDATE_REPO_UNSIGNED = "cannot-update-repo-unsigned"
ERROR_CANNOT_WRITE_REPO_CONFIG = "cannot-write-repo-config"
ERROR_CREATE_THREAD_FAILED = "create-thread-failed"
ERROR_DEP_RESOLUTION_FAILED = "dep-resolution-failed"
ERROR_FAILED_CONFIG_PARSING = "failed-config-parsing"
ERROR_FAILED_FINALISE = "failed-finalise"
ERROR_FAILED_INITIALIZATION = "failed-initialization"
ERROR_FILE_CONFLICTS = "file-conflicts"
ERROR_FILE_NOT_FOUND = "file-not-found"
ERROR_FILTER_INVALID = "filter-invalid"
ERROR_GPG_FAILURE = "gpg-failure"
ERROR_GROUP_LIST_INVALID = "group-list-invalid"
ERROR_GROUP_NOT_FOUND = "group-not-found"
ERROR_INCOMPATIBLE_ARCHITECTURE = "incompatible-architecture"
ERROR_INSTALL_ROOT_INVALID = "install-root-invalid"
ERROR_INTERNAL_ERROR = "internal-error"
ERROR_INVALID_PACKAGE_FILE = "invalid-package-file"
ERROR_LOCAL_INSTALL_FAILED = "local-install-failed"
ERROR_LOCK_REQUIRED = "lock-required"
ERROR_MEDIA_CHANGE_REQUIRED = "media-change-required"
ERROR_MISSING_GPG_SIGNATURE = "missing-gpg-signature"
ERROR_NOT_AUTHORIZED = "not-authorized"
ERROR_NOT_SUPPORTED = "not-supported"
ERROR_NO_CACHE = "no-cache"
ERROR_NO_DISTRO_UPGRADE_DATA = "no-distro-upgrade-data"
ERROR_NO_LICENSE_AGREEMENT = "no-license-agreement"
ERROR_NO_MORE_MIRRORS_TO_TRY = "no-more-mirrors-to-try"
ERROR_NO_NETWORK = "no-network"
ERROR_NO_PACKAGES_TO_UPDATE = "no-packages-to-update"
ERROR_NO_SPACE_ON_DEVICE = "no-space-on-device"
ERROR_OOM = "out-of-memory"
ERROR_PACKAGE_ALREADY_INSTALLED = "package-already-installed"
ERROR_PACKAGE_CONFLICTS = "package-conflicts"
ERROR_PACKAGE_CORRUPT = "package-corrupt"
ERROR_PACKAGE_DATABASE_CHANGED = "package-database-changed"
ERROR_PACKAGE_DOWNLOAD_FAILED = "package-download-failed"
ERROR_PACKAGE_FAILED_TO_BUILD = "package-failed-to-build"
ERROR_PACKAGE_FAILED_TO_CONFIGURE = "package-failed-to-configure"
ERROR_PACKAGE_FAILED_TO_INSTALL = "package-failed-to-install"
ERROR_PACKAGE_FAILED_TO_REMOVE = "package-failed-to-remove"
ERROR_PACKAGE_ID_INVALID = "package-id-invalid"
ERROR_PACKAGE_INSTALL_BLOCKED = "package-install-blocked"
ERROR_PACKAGE_NOT_FOUND = "package-not-found"
ERROR_PACKAGE_NOT_INSTALLED = "package-not-installed"
ERROR_PROCESS_KILL = "process-kill"
ERROR_PROVIDE_TYPE_NOT_SUPPORTED = "provide-type-not-supported"
ERROR_REPO_ALREADY_SET = "repo-already-set"
ERROR_REPO_CONFIGURATION_ERROR = "repo-configuration-error"
ERROR_REPO_NOT_AVAILABLE = "repo-not-available"
ERROR_REPO_NOT_FOUND = "repo-not-found"
ERROR_RESTRICTED_DOWNLOAD = "restricted-download"
ERROR_TRANSACTION_CANCELLED = "transaction-cancelled"
ERROR_TRANSACTION_ERROR = "transaction-error"
ERROR_UNFINISHED_TRANSACTION = "unfinished-transaction"
ERROR_UNKNOWN = "unknown"
ERROR_UPDATE_FAILED_DUE_TO_RUNNING_PROCESS = "failed-due-to-running-process"
ERROR_UPDATE_NOT_FOUND = "update-not-found"
EXIT_CANCELLED = "cancelled"
EXIT_CANCELLED_PRIORITY = "cancelled-priority"
EXIT_EULA_REQUIRED = "eula-required"
EXIT_FAILED = "failed"
EXIT_KEY_REQUIRED = "key-required"
EXIT_KILLED = "killed"
EXIT_MEDIA_CHANGE_REQUIRED = "media-change-required"
EXIT_NEED_UNTRUSTED = "need-untrusted"
EXIT_REPAIR_REQUIRED = "repair-required"
EXIT_SKIP_TRANSACTION = "skip-transaction"
EXIT_SUCCESS = "success"
EXIT_UNKNOWN = "unknown"
FILTER_APPLICATION = "application"
FILTER_ARCH = "arch"
FILTER_BASENAME = "basename"
FILTER_COLLECTIONS = "collections"
FILTER_DEVELOPMENT = "devel"
FILTER_DOWNLOADED = "downloaded"
FILTER_FREE = "free"
FILTER_GUI = "gui"
FILTER_INSTALLED = "installed"
FILTER_NEWEST = "newest"
FILTER_NONE = "none"
FILTER_NOT_APPLICATION = "~application"
FILTER_NOT_ARCH = "~arch"
FILTER_NOT_BASENAME = "~basename"
FILTER_NOT_COLLECTIONS = "~collections"
FILTER_NOT_DEVELOPMENT = "~devel"
FILTER_NOT_DOWNLOADED = "~downloaded"
FILTER_NOT_FREE = "~free"
FILTER_NOT_GUI = "~gui"
FILTER_NOT_INSTALLED = "~installed"
FILTER_NOT_NEWEST = "~newest"
FILTER_NOT_SOURCE = "~source"
FILTER_NOT_SUPPORTED = "~supported"
FILTER_NOT_VISIBLE = "~visible"
FILTER_SOURCE = "source"
FILTER_SUPPORTED = "supported"
FILTER_UNKNOWN = "unknown"
FILTER_VISIBLE = "visible"
GROUP_ACCESSIBILITY = "accessibility"
GROUP_ACCESSORIES = "accessories"
GROUP_ADMIN_TOOLS = "admin-tools"
GROUP_COLLECTIONS = "collections"
GROUP_COMMUNICATION = "communication"
GROUP_DESKTOP_DDE = "desktop-dde"
GROUP_DESKTOP_GNOME = "desktop-gnome"
GROUP_DESKTOP_KDE = "desktop-kde"
GROUP_DESKTOP_OTHER = "desktop-other"
GROUP_DESKTOP_XFCE = "desktop-xfce"
GROUP_DOCUMENTATION = "documentation"
GROUP_EDUCATION = "education"
GROUP_ELECTRONICS = "electronics"
GROUP_FONTS = "fonts"
GROUP_GAMES = "games"
GROUP_GRAPHICS = "graphics"
GROUP_INTERNET = "internet"
GROUP_LEGACY = "legacy"
GROUP_LOCALIZATION = "localization"
GROUP_MAPS = "maps"
GROUP_MULTIMEDIA = "multimedia"
GROUP_NETWORK = "network"
GROUP_NEWEST = "newest"
GROUP_OFFICE = "office"
GROUP_OTHER = "other"
GROUP_POWER_MANAGEMENT = "power-management"
GROUP_PROGRAMMING = "programming"
GROUP_PUBLISHING = "publishing"
GROUP_REPOS = "repos"
GROUP_SCIENCE = "science"
GROUP_SECURITY = "security"
GROUP_SERVERS = "servers"
GROUP_SYSTEM = "system"
GROUP_UNKNOWN = "unknown"
GROUP_VENDOR = "vendor"
GROUP_VIRTUALIZATION = "virtualization"
INFO_AVAILABLE = "available"
INFO_BLOCKED = "blocked"
INFO_BUGFIX = "bugfix"
INFO_CLEANUP = "cleanup"
INFO_COLLECTION_AVAILABLE = "collection-available"
INFO_COLLECTION_INSTALLED = "collection-installed"
INFO_CRITICAL = "critical"
INFO_DECOMPRESSING = "decompressing"
INFO_DOWNGRADE = "downgrade"
INFO_DOWNGRADING = "downgrading"
INFO_DOWNLOADING = "downloading"
INFO_ENHANCEMENT = "enhancement"
INFO_FINISHED = "finished"
INFO_IMPORTANT = "important"
INFO_INSTALL = "install"
INFO_INSTALLED = "installed"
INFO_INSTALLING = "installing"
INFO_LOW = "low"
INFO_NORMAL = "normal"
INFO_OBSOLETE = "obsolete"
INFO_OBSOLETING = "obsoleting"
INFO_PREPARING = "preparing"
INFO_REINSTALLING = "reinstalling"
INFO_REMOVE = "remove"
INFO_REMOVING = "removing"
INFO_SECURITY = "security"
INFO_TRUSTED = "trusted"
INFO_UNAVAILABLE = "unavailable"
INFO_UNKNOWN = "unknown"
INFO_UNTRUSTED = "untrusted"
INFO_UPDATING = "updating"
MEDIA_TYPE_CD = "cd"
MEDIA_TYPE_DISC = "disc"
MEDIA_TYPE_DVD = "dvd"
MEDIA_TYPE_UNKNOWN = "unknown"
NETWORK_MOBILE = "mobile"
NETWORK_OFFLINE = "offline"
NETWORK_ONLINE = "online"
NETWORK_UNKNOWN = "unknown"
NETWORK_WIFI = "wifi"
NETWORK_WIRED = "wired"
RESTART_APPLICATION = "application"
RESTART_NONE = "none"
RESTART_SECURITY_SESSION = "security-session"
RESTART_SECURITY_SYSTEM = "security-system"
RESTART_SESSION = "session"
RESTART_SYSTEM = "system"
RESTART_UNKNOWN = "unknown"
ROLE_ACCEPT_EULA = "accept-eula"
ROLE_CANCEL = "cancel"
ROLE_DEPENDS_ON = "depends-on"
ROLE_DOWNLOAD_PACKAGES = "download-packages"
ROLE_GET_CATEGORIES = "get-categories"
ROLE_GET_DETAILS = "get-details"
ROLE_GET_DETAILS_LOCAL = "get-details-local"
ROLE_GET_DISTRO_UPGRADES = "get-distro-upgrades"
ROLE_GET_FILES = "get-files"
ROLE_GET_FILES_LOCAL = "get-files-local"
ROLE_GET_OLD_TRANSACTIONS = "get-old-transactions"
ROLE_GET_PACKAGES = "get-packages"
ROLE_GET_REPO_LIST = "get-repo-list"
ROLE_GET_UPDATES = "get-updates"
ROLE_GET_UPDATE_DETAIL = "get-update-detail"
ROLE_INSTALL_FILES = "install-files"
ROLE_INSTALL_PACKAGES = "install-packages"
ROLE_INSTALL_SIGNATURE = "install-signature"
ROLE_REFRESH_CACHE = "refresh-cache"
ROLE_REMOVE_PACKAGES = "remove-packages"
ROLE_REPAIR_SYSTEM = "repair-system"
ROLE_REPO_ENABLE = "repo-enable"
ROLE_REPO_REMOVE = "repo-remove"
ROLE_REPO_SET_DATA = "repo-set-data"
ROLE_REQUIRED_BY = "required-by"
ROLE_RESOLVE = "resolve"
ROLE_SEARCH_DETAILS = "search-details"
ROLE_SEARCH_FILE = "search-file"
ROLE_SEARCH_GROUP = "search-group"
ROLE_SEARCH_NAME = "search-name"
ROLE_UNKNOWN = "unknown"
ROLE_UPDATE_PACKAGES = "update-packages"
ROLE_UPGRADE_SYSTEM = "upgrade-system"
ROLE_WHAT_PROVIDES = "what-provides"
SIGTYPE_GPG = "gpg"
SIGTYPE_UNKNOWN = "unknown"
STATUS_CANCEL = "cancel"
STATUS_CHECK_EXECUTABLE_FILES = "check-executable-files"
STATUS_CHECK_LIBRARIES = "check-libraries"
STATUS_CLEANUP = "cleanup"
STATUS_COMMIT = "commit"
STATUS_COPY_FILES = "copy-files"
STATUS_DEP_RESOLVE = "dep-resolve"
STATUS_DOWNLOAD = "download"
STATUS_DOWNLOAD_CHANGELOG = "download-changelog"
STATUS_DOWNLOAD_FILELIST = "download-filelist"
STATUS_DOWNLOAD_GROUP = "download-group"
STATUS_DOWNLOAD_PACKAGELIST = "download-packagelist"
STATUS_DOWNLOAD_REPOSITORY = "download-repository"
STATUS_DOWNLOAD_UPDATEINFO = "download-updateinfo"
STATUS_FINISHED = "finished"
STATUS_GENERATE_PACKAGE_LIST = "generate-package-list"
STATUS_INFO = "info"
STATUS_INSTALL = "install"
STATUS_LOADING_CACHE = "loading-cache"
STATUS_OBSOLETE = "obsolete"
STATUS_QUERY = "query"
STATUS_REFRESH_CACHE = "refresh-cache"
STATUS_REMOVE = "remove"
STATUS_REPACKAGING = "repackaging"
STATUS_REQUEST = "request"
STATUS_RUNNING = "running"
STATUS_RUN_HOOK = "run-hook"
STATUS_SCAN_APPLICATIONS = "scan-applications"
STATUS_SCAN_PROCESS_LIST = "scan-process-list"
STATUS_SETUP = "setup"
STATUS_SIG_CHECK = "sig-check"
STATUS_TEST_COMMIT = "test-commit"
STATUS_UNKNOWN = "unknown"
STATUS_UPDATE = "update"
STATUS_WAIT = "wait"
STATUS_WAITING_FOR_AUTH = "waiting-for-auth"
STATUS_WAITING_FOR_LOCK = "waiting-for-lock"
TRANSACTION_FLAG_ALLOW_DOWNGRADE = "allow-downgrade"
TRANSACTION_FLAG_ALLOW_REINSTALL = "allow-reinstall"
TRANSACTION_FLAG_JUST_REINSTALL = "just-reinstall"
TRANSACTION_FLAG_NONE = "none"
TRANSACTION_FLAG_ONLY_DOWNLOAD = "only-download"
TRANSACTION_FLAG_ONLY_TRUSTED = "only-trusted"
TRANSACTION_FLAG_SIMULATE = "simulate"
UPDATE_STATE_STABLE = "stable"
UPDATE_STATE_TESTING = "testing"
UPDATE_STATE_UNKNOWN = "unknown"
UPDATE_STATE_UNSTABLE = "unstable"
UPGRADE_KIND_COMPLETE = "complete"
UPGRADE_KIND_DEFAULT = "default"
UPGRADE_KIND_MINIMAL = "minimal"
UPGRADE_KIND_UNKNOWN = "unknown"



# Licensed under the GNU General Public License Version 2
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.

# Copyright (C) 2007 Tim Lauridsen <timlau@fedoraproject.org>
# Copyright (C) 2007-2010 Richard Hughes <richard@hughsie.com>
#
# This file contain the base classes to implement a PackageKit python backend
#


# imports
from __future__ import print_function

import sys
import traceback
import os.path

from .enums import *

PACKAGE_IDS_DELIM = '&'
FILENAME_DELIM = '|'

MAXUINT64 = (1 << 64) - 1

def _to_unicode(txt, encoding='utf-8'):
    if isinstance(txt, str):
        if not isinstance(txt, str):
            txt = str(txt, encoding, errors='replace')
    return txt

def _to_utf8(txt, errors='replace'):
    if isinstance(txt, str):
        return txt
    if isinstance(txt, unicode):
        return txt.encode('utf-8', errors=errors)
    return str(txt)

class PkError(Exception):
    def __init__(self, code, details):
        self.code = code
        self.details = details
    def __str__(self):
        return repr("%s: %s" % (self.code, self.details))

class PackageKitBaseBackend:

    def __init__(self, cmds):
        # Setup a custom exception handler
        installExceptionHandler(self)
        self.cmds = cmds
        self._locked = False
        self.lang = "C"
        self.has_network = False
        self.uid = 0
        self.background = False
        self.interactive = False
        self.cache_age = 0
        self.percentage_old = 0

        # try to get LANG
        try:
            self.lang = os.environ['LANG']
        except KeyError as e:
            print("Error: No LANG envp")

        # try to get NETWORK state
        try:
            if os.environ['NETWORK'] == 'TRUE':
                self.has_network = True
        except KeyError as e:
            print("Error: No NETWORK envp")

        # try to get UID of running user
        try:
            self.uid = int(os.environ['UID'])
        except KeyError as e:
            print("Error: No UID envp")

        # try to get BACKGROUND state
        try:
            if os.environ['BACKGROUND'] == 'TRUE':
                self.background = True
        except KeyError as e:
            print("Error: No BACKGROUND envp")

        # try to get INTERACTIVE state
        try:
            if os.environ['INTERACTIVE'] == 'TRUE':
                self.interactive = True
        except KeyError as e:
            print("Error: No INTERACTIVE envp")

        # try to get CACHE_AGE state
        try:
            self.cache_age = int(os.environ['CACHE_AGE'])
        except KeyError as e:
            pass

    def doLock(self):
        ''' Generic locking, overide and extend in child class'''
        self._locked = True

    def unLock(self):
        ''' Generic unlocking, overide and extend in child class'''
        self._locked = False

    def isLocked(self):
        return self._locked

    def percentage(self, percent=None):
        '''
        Write progress percentage
        @param percent: Progress percentage (int preferred)
        '''
        if percent == None:
            sys.stdout.write(_to_utf8("no-percentage-updates\n"))
        elif percent == 0 or percent > self.percentage_old:
            sys.stdout.write(_to_utf8("percentage\t%i\n" % percent))
            self.percentage_old = percent
        sys.stdout.flush()

    def speed(self, bps=0):
        '''
        Write progress speed
        @param bps: Progress speed (int, bytes per second)
        '''
        sys.stdout.write(_to_utf8("speed\t%i\n" % bps))
        sys.stdout.flush()

    def item_progress(self, package_id, status, percent=None):
        '''
        send 'itemprogress' signal
        @param package_id: The package ID name, e.g. openoffice-clipart;2.6.22;ppc64;fedora
        @param percent: percentage of the current item (int preferred)
        '''
        sys.stdout.write(_to_utf8("item-progress\t%s\t%s\t%i\n" % (package_id, status, percent)))
        sys.stdout.flush()

    def error(self, err, description, exit=True):
        '''
        send 'error'
        @param err: Error Type (ERROR_NO_NETWORK, ERROR_NOT_SUPPORTED, ERROR_INTERNAL_ERROR)
        @param description: Error description
        @param exit: exit application with rc = 1, if true
        '''
        # unlock before we emit if we are going to exit
        if exit and self.isLocked():
            self.unLock()

        # this should be fast now
        sys.stdout.write(_to_utf8("error\t%s\t%s\n" % (err, description)))
        sys.stdout.flush()
        if exit:
            # Paradoxically, we don't want to print "finished" to stdout here.
            # Python takes an _enormous_ amount of time to exit, and leaves a
            # huge race when you try to change a dispatcher because of an error.
            #
            # Leave PackageKit to clean up for us in this case.
            sys.exit(254)

    def message(self, typ, msg):
        '''
        send 'message' signal
        @param typ: MESSAGE_BROKEN_MIRROR
        '''
        sys.stdout.write(_to_utf8("message\t%s\t%s\n" % (typ, msg)))
        sys.stdout.flush()

    def package(self, package_id, status, summary):
        '''
        send 'package' signal
        @param info: the enumerated INFO_* string
        @param package_id: The package ID name, e.g. openoffice-clipart;2.6.22;ppc64;fedora
        @param summary: The package Summary
        '''
        sys.stdout.write(_to_utf8("package\t%s\t%s\t%s\n" % (status, package_id, summary)))
        sys.stdout.flush()

    def media_change_required(self, mtype, id, text):
        '''
        send 'media-change-required' signal
        @param mtype: the enumerated MEDIA_TYPE_* string
        @param id: the localised label of the media
        @param text: the localised text describing the media
        '''
        sys.stdout.write(_to_utf8("media-change-required\t%s\t%s\t%s\n" % (mtype, id, text)))
        sys.stdout.flush()

    def distro_upgrade(self, dtype, name, summary):
        '''
        send 'distro-upgrade' signal
        @param dtype: the enumerated DISTRO_UPGRADE_* string
        @param name: The distro name, e.g. "fedora-9"
        @param summary: The localised distribution name and description
        '''
        sys.stdout.write(_to_utf8("distro-upgrade\t%s\t%s\t%s\n" % (dtype, name, summary)))
        sys.stdout.flush()

    def status(self, state):
        '''
        send 'status' signal
        @param state: STATUS_DOWNLOAD, STATUS_INSTALL, STATUS_UPDATE, STATUS_REMOVE, STATUS_WAIT
        '''
        sys.stdout.write(_to_utf8("status\t%s\n" % state))
        sys.stdout.flush()

    def repo_detail(self, repoid, name, state):
        '''
        send 'repo-detail' signal
        @param repoid: The repo id tag
        @param state: false is repo is disabled else true.
        '''
        sys.stdout.write(_to_utf8("repo-detail\t%s\t%s\t%s\n" % (repoid, name, _bool_to_string(state))))
        sys.stdout.flush()

    def data(self, data):
        '''
        send 'data' signal:
        @param data:  The current worked on package
        '''
        sys.stdout.write(_to_utf8("data\t%s\n" % data))
        sys.stdout.flush()

    def details(self, package_id, summary, package_license, group, desc, url, bytes: int | None = None, download_bytes: int | None = None):
        '''
        Send 'details' signal
        @param package_id: The package ID name, e.g. openoffice-clipart;2.6.22;ppc64;fedora
        @param summary: The package summary
        @param package_license: The license of the package
        @param group: The enumerated group
        @param desc: The multi line package description
        @param url: The upstream project homepage
        @param bytes: The size of the package, in bytes
        @param download_bytes: The download size of the package, in bytes
        '''

        # Ensure sizes report as 'unknown' in pkgcli if not reported
        if bytes is None:
            bytes = MAXUINT64
        if download_bytes is None:
            download_bytes = MAXUINT64

        sys.stdout.write(_to_utf8("details\t%s\t%s\t%s\t%s\t%s\t%s\t%ld\t%ld\n" % (package_id, summary, package_license, group, desc, url, bytes, download_bytes)))
        sys.stdout.flush()

    def files(self, package_id, file_list):
        '''
        Send 'files' signal
        @param file_list: List of the files in the package, separated by ';'
        '''
        sys.stdout.write(_to_utf8("files\t%s\t%s\n" % (package_id, file_list)))
        sys.stdout.flush()

    def category(self, parent_id, cat_id, name, summary, icon):
        '''
        Send 'category' signal
        parent_id : A parent id, e.g. "admin" or "" if there is no parent
        cat_id    : a unique category id, e.g. "admin;network"
        name      : a verbose category name in current locale.
        summery   : a summary of the category in current locale.
        icon      : an icon name to represent the category
        '''
        sys.stdout.write(_to_utf8("category\t%s\t%s\t%s\t%s\t%s\n" % (parent_id, cat_id, name, summary, icon)))
        sys.stdout.flush()

    def finished(self):
        '''
        Send 'finished' signal
        '''
        sys.stdout.write(_to_utf8("finished\n"))
        sys.stdout.flush()

    def update_detail(self, package_id, updates, obsoletes, vendor_url, bugzilla_url, cve_url, restart, update_text, changelog, state, issued, updated):
        '''
        Send 'updatedetail' signal
        @param package_id: The package ID name, e.g. openoffice-clipart;2.6.22;ppc64;fedora
        @param updates:
        @param obsoletes:
        @param vendor_url:
        @param bugzilla_url:
        @param cve_url:
        @param restart:
        @param update_text:
        @param changelog:
        @param state:
        @param issued:
        @param updated:
        '''
        sys.stdout.write(_to_utf8("updatedetail\t%s\t%s\t%s\t%s\t%s\t%s\t%s\t%s\t%s\t%s\t%s\t%s\n" % (package_id, updates, obsoletes, vendor_url, bugzilla_url, cve_url, restart, update_text, changelog, state, issued, updated)))
        sys.stdout.flush()

    def require_restart(self, restart_type, details):
        '''
        Send 'requirerestart' signal
        @param restart_type: RESTART_SYSTEM, RESTART_APPLICATION, RESTART_SESSION
        @param details: Optional details about the restart
        '''
        sys.stdout.write(_to_utf8("requirerestart\t%s\t%s\n" % (restart_type, details)))
        sys.stdout.flush()

    def allow_cancel(self, allow):
        '''
        send 'allow-cancel' signal:
        @param allow:  Allow the current process to be aborted.
        '''
        if allow:
            data = 'true'
        else:
            data = 'false'
        sys.stdout.write(_to_utf8("allow-cancel\t%s\n" % data))
        sys.stdout.flush()

    def repo_signature_required(self, package_id, repo_name, key_url, key_userid, key_id, key_fingerprint, key_timestamp, sig_type):
        '''
        send 'repo-signature-required' signal:
        @param package_id:      Id of the package needing a signature
        @param repo_name:       Name of the repository
        @param key_url:         URL which the user can use to verify the key
        @param key_userid:      Key userid
        @param key_id:          Key ID
        @param key_fingerprint: Full key fingerprint
        @param key_timestamp:   Key timestamp
        @param sig_type:        Key type (GPG)
        '''
        sys.stdout.write(_to_utf8("repo-signature-required\t%s\t%s\t%s\t%s\t%s\t%s\t%s\t%s\n" % (
            package_id, repo_name, key_url, key_userid, key_id, key_fingerprint, key_timestamp, sig_type
            )))
        sys.stdout.flush()

    def eula_required(self, eula_id, package_id, vendor_name, license_agreement):
        '''
        send 'eula-required' signal:
        @param eula_id:         Id of the EULA
        @param package_id:      Id of the package needing a signature
        @param vendor_name:     Name of the vendor that wrote the EULA
        @param license_agreement: The license text
        '''
        sys.stdout.write(_to_utf8("eula-required\t%s\t%s\t%s\t%s\n" % (
            eula_id, package_id, vendor_name, license_agreement
            )))
        sys.stdout.flush()

#
# Backend Action Methods
#

    def search_name(self, filters, values):
        '''
        Implement the {backend}-search-name functionality
        Needed to be implemented in a sub class
        '''
        self.error(ERROR_NOT_SUPPORTED, "This function is not implemented in this backend", exit=False)

    def search_details(self, filters, values):
        '''
        Implement the {backend}-search-details functionality
        Needed to be implemented in a sub class
        '''
        self.error(ERROR_NOT_SUPPORTED, "This function is not implemented in this backend", exit=False)

    def search_group(self, filters, values):
        '''
        Implement the {backend}-search-group functionality
        Needed to be implemented in a sub class
        '''
        self.error(ERROR_NOT_SUPPORTED, "This function is not implemented in this backend", exit=False)

    def search_file(self, filters, values):
        '''
        Implement the {backend}-search-file functionality
        Needed to be implemented in a sub class
        '''
        self.error(ERROR_NOT_SUPPORTED, "This function is not implemented in this backend", exit=False)

    def get_update_detail(self, package_ids):
        '''
        Implement the {backend}-get-update-detail functionality
        Needed to be implemented in a sub class
        '''
        self.error(ERROR_NOT_SUPPORTED, "This function is not implemented in this backend", exit=False)

    def depends_on(self, filters, package_ids, recursive):
        '''
        Implement the {backend}-depends-on functionality
        Needed to be implemented in a sub class
        '''
        self.error(ERROR_NOT_SUPPORTED, "This function is not implemented in this backend", exit=False)

    def get_packages(self, filters):
        '''
        Implement the {backend}-get-packages functionality
        Needed to be implemented in a sub class
        '''
        self.error(ERROR_NOT_SUPPORTED, "This function is not implemented in this backend", exit=False)

    def required_by(self, filters, package_ids, recursive):
        '''
        Implement the {backend}-required-by functionality
        Needed to be implemented in a sub class
        '''
        self.error(ERROR_NOT_SUPPORTED, "This function is not implemented in this backend", exit=False)

    def what_provides(self, filters, provides_type, values):
        '''
        Implement the {backend}-what-provides functionality
        Needed to be implemented in a sub class
        '''
        self.error(ERROR_NOT_SUPPORTED, "This function is not implemented in this backend", exit=False)

    def upgrade_system(self, distro_id):
        '''
        Implement the {backend}-update-system functionality
        Needed to be implemented in a sub class
        '''
        self.error(ERROR_NOT_SUPPORTED, "This function is not implemented in this backend", exit=False)

    def refresh_cache(self, force):
        '''
        Implement the {backend}-refresh_cache functionality
        Needed to be implemented in a sub class
        '''
        self.error(ERROR_NOT_SUPPORTED, "This function is not implemented in this backend", exit=False)

    def install_packages(self, transaction_flags, package_ids):
        '''
        Implement the {backend}-install functionality
        Needed to be implemented in a sub class
        '''
        self.error(ERROR_NOT_SUPPORTED, "This function is not implemented in this backend", exit=False)

    def install_signature(self, sigtype, key_id, package_id):
        '''
        Implement the {backend}-install-signature functionality
        Needed to be implemented in a sub class
        '''
        self.error(ERROR_NOT_SUPPORTED, "This function is not implemented in this backend", exit=False)

    def install_files(self, transaction_flags, inst_files):
        '''
        Implement the {backend}-install_files functionality
        Install the package containing the inst_file file
        Needed to be implemented in a sub class
        '''
        self.error(ERROR_NOT_SUPPORTED, "This function is not implemented in this backend", exit=False)

    def resolve(self, filters, values):
        '''
        Implement the {backend}-resolve functionality
        Needed to be implemented in a sub class
        '''
        self.error(ERROR_NOT_SUPPORTED, "This function is not implemented in this backend", exit=False)

    def remove_packages(self, transaction_flags, package_ids, allowdep, autoremove):
        '''
        Implement the {backend}-remove functionality
        Needed to be implemented in a sub class
        '''
        self.error(ERROR_NOT_SUPPORTED, "This function is not implemented in this backend", exit=False)

    def update_packages(self, transaction_flags, package_ids):
        '''
        Implement the {backend}-update functionality
        Needed to be implemented in a sub class
        '''
        self.error(ERROR_NOT_SUPPORTED, "This function is not implemented in this backend", exit=False)

    def get_details(self, package_ids):
        '''
        Implement the {backend}-get-details functionality
        Needed to be implemented in a sub class
        '''
        self.error(ERROR_NOT_SUPPORTED, "This function is not implemented in this backend", exit=False)

    def get_details_local(self, files):
        '''
        Implement the {backend}-get-details-local functionality
        Needed to be implemented in a sub class
        '''
        self.error(ERROR_NOT_SUPPORTED, "This function is not implemented in this backend", exit=False)

    def get_files(self, package_ids):
        '''
        Implement the {backend}-get-files functionality
        Needed to be implemented in a sub class
        '''
        self.error(ERROR_NOT_SUPPORTED, "This function is not implemented in this backend", exit=False)

    def get_updates(self, filters):
        '''
        Implement the {backend}-get-updates functionality
        Needed to be implemented in a sub class
        '''
        self.error(ERROR_NOT_SUPPORTED, "This function is not implemented in this backend", exit=False)

    def get_distro_upgrades(self):
        '''
        Implement the {backend}-get-distro-upgrades functionality
        Needed to be implemented in a sub class
        '''
        self.error(ERROR_NOT_SUPPORTED, "This function is not implemented in this backend", exit=False)

    def repo_enable(self, repoid, enable):
        '''
        Implement the {backend}-repo-enable functionality
        Needed to be implemented in a sub class
        '''
        self.error(ERROR_NOT_SUPPORTED, "This function is not implemented in this backend", exit=False)

    def repo_set_data(self, repoid, parameter, value):
        '''
        Implement the {backend}-repo-set-data functionality
        Needed to be implemented in a sub class
        '''
        self.error(ERROR_NOT_SUPPORTED, "This function is not implemented in this backend", exit=False)

    def get_repo_list(self, filters):
        '''
        Implement the {backend}-get-repo-list functionality
        Needed to be implemented in a sub class
        '''
        self.error(ERROR_NOT_SUPPORTED, "This function is not implemented in this backend", exit=False)

    def repo_signature_install(self, package_id):
        '''
        Implement the {backend}-repo-signature-install functionality
        Needed to be implemented in a sub class
        '''
        self.error(ERROR_NOT_SUPPORTED, "This function is not implemented in this backend", exit=False)

    def download_packages(self, directory, package_ids):
        '''
        Implement the {backend}-download-packages functionality
        Needed to be implemented in a sub class
        '''
        self.error(ERROR_NOT_SUPPORTED, "This function is not implemented in this backend", exit=False)

    def set_locale(self, code):
        '''
        Implement the {backend}-set-locale functionality
        Needed to be implemented in a sub class
        '''
        self.error(ERROR_NOT_SUPPORTED, "This function is not implemented in this backend", exit=False)

    def get_categories(self):
        '''
        Implement the {backend}-get-categories functionality
        Needed to be implemented in a sub class
        '''
        self.error(ERROR_NOT_SUPPORTED, "This function is not implemented in this backend", exit=False)

    def repair_system(self, transaction_flags):
        '''
        Implement the {backend}-repair-system functionality
        Needed to be implemented in a sub class
        '''
        self.error(ERROR_NOT_SUPPORTED, "This function is not implemented in this backend",
                   exit=False)

    def customTracebackHandler(self, tb):
        '''
        Custom Traceback Handler
        this is called by the ExceptionHandler
        return True if the exception is handled in the method.
        return False if to do the default action an signal an error
        to packagekit.
        Overload this method if you what handle special Tracebacks
        '''
        return False

    def run_command(self):
        '''
        interprete the command from the calling args (self.cmds)
        '''
        fname = os.path.split(self.cmds[0])[1]
        cmd = fname.split('.')[0] # get the helper filename wo ext
        args = self.cmds[1:]
        self.dispatch_command(cmd, args)

    def dispatch_command(self, cmd, args):
        if cmd == 'download-packages':
            directory = args[0]
            package_ids = args[1].split(PACKAGE_IDS_DELIM)
            self.download_packages(directory, package_ids)
            self.finished()
        elif cmd == 'depends-on':
            filters = args[0].split(';')
            package_ids = args[1].split(PACKAGE_IDS_DELIM)
            recursive = _text_to_bool(args[2])
            self.depends_on(filters, package_ids, recursive)
            self.finished()
        elif cmd == 'get-details':
            package_ids = args[0].split(PACKAGE_IDS_DELIM)
            self.get_details(package_ids)
            self.finished()
        elif cmd == 'get-details-local':
            files = args[0].split(PACKAGE_IDS_DELIM)
            self.get_details_local(files)
            self.finished()
        elif cmd == 'get-files':
            package_ids = args[0].split(PACKAGE_IDS_DELIM)
            self.get_files(package_ids)
            self.finished()
        elif cmd == 'get-packages':
            filters = args[0].split(';')
            self.get_packages(filters)
            self.finished()
        elif cmd == 'get-repo-list':
            filters = args[0].split(';')
            self.get_repo_list(filters)
            self.finished()
        elif cmd == 'required-by':
            filters = args[0].split(';')
            package_ids = args[1].split(PACKAGE_IDS_DELIM)
            recursive = _text_to_bool(args[2])
            self.required_by(filters, package_ids, recursive)
            self.finished()
        elif cmd == 'get-update-detail':
            package_ids = args[0].split(PACKAGE_IDS_DELIM)
            self.get_update_detail(package_ids)
            self.finished()
        elif cmd == 'get-distro-upgrades':
            self.get_distro_upgrades()
            self.finished()
        elif cmd == 'get-updates':
            filters = args[0].split(';')
            self.get_updates(filters)
            self.finished()
        elif cmd == 'install-files':
            transaction_flags = args[0].split(';')
            files_to_inst = args[1].split(FILENAME_DELIM)
            self.install_files(transaction_flags, files_to_inst)
            self.finished()
        elif cmd == 'install-packages':
            transaction_flags = args[0].split(';')
            package_ids = args[1].split(PACKAGE_IDS_DELIM)
            self.install_packages(transaction_flags, package_ids)
            self.finished()
        elif cmd == 'install-signature':
            sigtype = args[0]
            key_id = args[1]
            package_id = args[2]
            self.install_signature(sigtype, key_id, package_id)
            self.finished()
        elif cmd == 'refresh-cache':
            force = _text_to_bool(args[0])
            self.refresh_cache(force)
            self.finished()
        elif cmd == 'remove-packages':
            transaction_flags = args[0].split(';')
            package_ids = args[1].split(PACKAGE_IDS_DELIM)
            allowdeps = _text_to_bool(args[2])
            autoremove = _text_to_bool(args[3])
            self.remove_packages(transaction_flags, package_ids, allowdeps, autoremove)
            self.finished()
        elif cmd == 'repo-enable':
            repoid = args[0]
            state = _text_to_bool(args[1])
            self.repo_enable(repoid, state)
            self.finished()
        elif cmd == 'repo-set-data':
            repoid = args[0]
            para = args[1]
            value = args[2]
            self.repo_set_data(repoid, para, value)
            self.finished()
        elif cmd == 'resolve':
            filters = args[0].split(';')
            package_ids = args[1].split(PACKAGE_IDS_DELIM)
            self.resolve(filters, package_ids)
            self.finished()
        elif cmd == 'search-details':
            filters = args[0].split(';')
            values = _to_unicode(args[1]).split(PACKAGE_IDS_DELIM)
            self.search_details(filters, values)
            self.finished()
        elif cmd == 'search-file':
            filters = args[0].split(';')
            values = args[1].split(PACKAGE_IDS_DELIM)
            self.search_file(filters, values)
            self.finished()
        elif cmd == 'search-group':
            filters = args[0].split(';')
            values = args[1].split(PACKAGE_IDS_DELIM)
            self.search_group(filters, values)
            self.finished()
        elif cmd == 'search-name':
            filters = args[0].split(';')
            values = _to_unicode(args[1]).split(PACKAGE_IDS_DELIM)
            self.search_name(filters, values)
            self.finished()
        elif cmd == 'signature-install':
            package = args[0]
            self.repo_signature_install(package)
            self.finished()
        elif cmd == 'update-packages':
            transaction_flags = args[0].split(';')
            package_ids = args[1].split(PACKAGE_IDS_DELIM)
            self.update_packages(transaction_flags, package_ids)
            self.finished()
        elif cmd == 'what-provides':
            filters = args[0].split(';')
            provides_type = args[1]
            values = _to_unicode(args[2]).split(PACKAGE_IDS_DELIM)
            self.what_provides(filters, provides_type, values)
            self.finished()
        elif cmd == 'set-locale':
            code = args[0]
            self.set_locale(code)
            self.finished()
        elif cmd == 'get-categories':
            self.get_categories()
            self.finished()
        elif cmd == 'upgrade-system':
            self.upgrade_system(args[0])
            self.finished()
        elif cmd == 'repair-system':
            self.repair_system(args[0])
            self.finished()
        else:
            errmsg = "command '%s' is not known" % cmd
            self.error(ERROR_INTERNAL_ERROR, errmsg, exit=False)
            self.finished()

    def dispatcher(self, args):
        if len(args) > 0:
            self.dispatch_command(args[0], args[1:])
        while True:
            try:
                line = sys.stdin.readline().strip('\n')
            except IOError as e:
                self.error(ERROR_TRANSACTION_CANCELLED, 'could not read from stdin: %s' % str(e))
            except KeyboardInterrupt as e:
                self.error(ERROR_PROCESS_KILL, 'process was killed by ctrl-c: %s' % str(e))
            if not line or line == 'exit':
                break
            args = line.split('\t')
            self.dispatch_command(args[0], args[1:])

        # unlock backend and exit with success
        if self.isLocked():
            self.unLock()
        sys.exit(0)


def format_string(text, encoding='utf-8'):
    '''
    Format a string to be used on stdout for communication with the daemon.
    '''
    if not isinstance(text, str):
        text = str(text, encoding, errors='replace')
    return text.replace("\n", ";")

def _text_to_bool(text):
    '''Convert a string to a boolean value.'''
    if text.lower() in ["yes", "true"]:
        return True
    return False

def _bool_to_string(value):
    if value:
        return "true"
    return "false"

def get_package_id(name, version, arch, data):
    """Returns a package id."""
    return ";".join((name, version, arch, data))

def split_package_id(id):
    """
    Returns a tuple with the name, version, arch and data component of a
    package id.
    """
    return id.split(";", 4)

def exceptionHandler(typ, value, tb, base):
    # Restore original exception handler
    sys.excepthook = sys.__excepthook__
    # Call backend custom Traceback handler
    if not base.customTracebackHandler(typ):
        etb = traceback.extract_tb(tb)
        errmsg = 'Error Type: %s;' % str(typ)
        errmsg += 'Error Value: %s;' % str(value)
        for tub in etb:
            f, l, m, c = tub # file, lineno, function, codeline
            errmsg += '  File : %s, line %s, in %s;' % (f, str(l), m)
            errmsg += '    %s;' % c
        # send the traceback to PackageKit
        base.error(ERROR_INTERNAL_ERROR, errmsg, exit=True)

def installExceptionHandler(base):
    sys.excepthook = lambda typ, value, tb: exceptionHandler(typ, value, tb, base)

